//===- AnalysisPass.cpp - AutoJIT Analysis Pass -----------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "autojit-analysis"

#include <map>
#include <set>
#include <vector>

#include "BranchMerge.h"
#include "AnalysisPass.h"

#include "llvm/BasicBlock.h"
#include "llvm/Metadata.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Local.h"
#include "llvm/Support/CFG.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/ADT/DepthFirstIterator.h"
using namespace llvm;

namespace autojit {

  bool AnalysisPass::runOnFunction(Function& f)
  {
    BranchMerge& bmInfo = getAnalysis<BranchMerge>();
    LoopInfo& loopInfo = getAnalysis<LoopInfo>();

    std::map<Instruction*,unsigned int> vars;
    for (inst_iterator instIter = inst_begin(f);
         instIter != inst_end(f); ++instIter) {
      if (CallInst* callInst = dyn_cast<CallInst>(&*instIter)) {
        if (Function* callee = callInst->getCalledFunction()) {
          if (callee->getName().startswith("__autojit__cvar_")) {
            if (!callee->isDeclaration()) {
              errs() << "ERR: Annotation function has body.\n";
              continue;
            }
            const FunctionType* fnTy = callee->getFunctionType();
            if (1 != fnTy->getNumParams() || fnTy->isVarArg() ||
                fnTy->getParamType(0) != fnTy->getReturnType()) {
              errs() << "ERR: Annotation function has bad type.\n";
              continue;
            }
            vars.insert(std::pair<Instruction*,unsigned int>(callInst,InstPtr));
            mAnnotations.insert(callInst);
            continue;
          }
        }
      }
    }

    // Follow dataflows into function
    unsigned int numVars;
    do {
      numVars = vars.size();
      for (std::map<Instruction*,unsigned int>::iterator valIter = vars.begin(),
           valIterE = vars.end(); valIter != valIterE; ++valIter) {
        Instruction* const srcInst = valIter->first;
        for (Value::use_iterator useIter = srcInst->use_begin(),
             useIterE = srcInst->use_end();
             useIter != useIterE; ++useIter) {
          Instruction* inst = cast<Instruction>(useIter);
          unsigned int op = inst->getOpcode();
          if (Instruction::Load != op && Instruction::Call != op &&
              Instruction::Invoke != op) {
            unsigned int incr =
              valIter->first->getParent() != inst->getParent();
            vars.insert(std::pair<Instruction*,unsigned int>(
              inst, valIter->second+incr));
          }
        }
      }
    } while (vars.size() != numVars);

    // Find interpreter loop
    Loop* interpLoop = NULL;
    unsigned int phiDistance = ~0;
    for (std::map<Instruction*,unsigned int>::iterator varIter = vars.begin(),
         varIterE = vars.end(); varIter != varIterE; ++varIter) {
      if (PHINode* phi = dyn_cast<PHINode>(varIter->first)) {
        if (phi->getParent() == mHead) {
          mCtrlVars.insert(phi);
        }
        else {
          // Test for loop header
          Loop* loop = loopInfo.getLoopFor(phi->getParent());
          if (!loop || (loop->getHeader() != phi->getParent())) {
            continue;
          }

          // Test for distance
          if (varIter->second > phiDistance) {
            continue;
          }

          // Clear the set and insert a phi node from the new candidate block
          interpLoop = loop;
          phiDistance = varIter->second;
          mHead = phi->getParent();
          mCtrlVars.clear();
          mCtrlVars.insert(phi);
        }
      }
    }
    if (!mHead) {
      return false;
    }

    // Find enclosing loops
    std::set<Loop*> outerLoops, innerLoops;
    for (Loop* loop = loopInfo.getLoopFor(mHead); loop;
         loop = loop->getParentLoop()) {
      outerLoops.insert(loop);

      // Find inner loops
      for (Loop::iterator loopIter = loop->begin(), loopIterE = loop->end();
           loopIter != loopIterE; ++loopIter) {
        if (!outerLoops.count(*loopIter)) {
          innerLoops.insert(*loopIter);
        }
      }
    }

    // Find remaining innner loops
    for (Loop::iterator loopIter = loopInfo.begin(), loopIterE = loopInfo.end();
         loopIter != loopIterE; ++loopIter) {
      Loop* loop = *loopIter;
      if (!outerLoops.count(loop) && mBody.count(loop->getHeader())) {
        innerLoops.insert(loop);
      }
    }

    // Add the head phi instructions for non-control data-flows to the bad set
//    std::set<Instruction*> badDataflows;
    for (BasicBlock::iterator instIter = mHead->begin(),
         instIterE = mHead->end(); instIter != instIterE; ++instIter) {
      if (PHINode* phi = dyn_cast<PHINode>(&*instIter)) {
        if (!mCtrlVars.count(phi)) {
          mRtVars.insert(phi);
//          badDataflows.insert(phi);
        }
      }
    }

    // Enumerate blocks reachable from the head
    for (df_ext_iterator<BasicBlock*,std::set<BasicBlock*> > dfIter =
         df_ext_begin(mHead, mBody), dfIterE = df_ext_end(mHead, mBody);
         dfIter != dfIterE; ++dfIter) {
      // Backwards: Scan function returns
      TerminatorInst* termInst = (*dfIter)->getTerminator();
      if (!termInst->getNumSuccessors()) {
        mCtrlFlowVals.insert(termInst);
      }
    }

    // Prepare list of body blocks
    for (Function::iterator blkIter = f.begin(),
         blkIterE = f.end(); blkIter != blkIterE; ++blkIter) {
      BasicBlock* block = &*blkIter;
      if (mBody.count(block)) {
         mBodyList.push_back(block);
      }
    }
    while (mBodyList.front() != mHead) {
      mBodyList.push_back(mBodyList.front());
      mBodyList.pop_front();
    }

    // Loop over annotated PHI instructions
    FwdScan goodDataflows;
    for (std::set<PHINode*>::iterator phiIter = mCtrlVars.begin(),
         phiIterE = mCtrlVars.end(); phiIter != phiIterE; ++phiIter) {
      PHINode* phi = *phiIter;

      // Forwards: Add annotated dataflows to the good set
      goodDataflows.insert(FwdPair(phi, InstPtr));

      // Backwards: Scan the tails of instptr dataflows
      for (unsigned int i = 0, iE = phi->getNumIncomingValues(); i < iE; i++) {
        if (mBody.count(phi->getIncomingBlock(i))) {
          if (Instruction* inst =
                dyn_cast<Instruction>(phi->getIncomingValue(i))) {
            mCtrlFlowVals.insert(inst);
          }
        }
      }
    }

    // Iterate over body instructions
    std::set<Instruction*> scanFront;
    for (std::set<BasicBlock*>::iterator blockIter = mBody.begin(),
         blockIterE = mBody.end(); blockIter != blockIterE; ++blockIter) {
      BasicBlock* block = *blockIter;
      for (BasicBlock::iterator instIter = block->begin(),
           instIterE = block->end(); instIter != instIterE; ++instIter) {
        Instruction* inst = &*instIter;

        // Account for invariants from outside the body
        if (!isa<PHINode>(inst)) {
          for (Instruction::op_iterator opIter = inst->op_begin(),
               opIterE = inst->op_end(); opIter != opIterE; ++opIter) {
            Value* opValue = *opIter;
            if (isa<Argument>(opValue)) {
              mInvars.insert(opValue);
            }
            else if (Instruction* opInst = dyn_cast<Instruction>(opValue)) {
              if (!mBody.count(opInst->getParent())) {
                mInvars.insert(opInst);
              }
            }
          }
        }

        // Add untainted dataflows to front of forwards scan
        if (!mRtVars.count(dyn_cast<PHINode>(inst))) {
          scanFront.insert(inst);
        }

        // Record unconditional branches
        if (TerminatorInst* termInst = dyn_cast<TerminatorInst>(inst)) {
          if (1 == termInst->getNumSuccessors()) {
            mCtrlFlowVals.insert(termInst);
          }
        }
      }
    }

    // Forward propagate good dataflows
    while (!scanFront.empty()) {
      for (std::set<Instruction*>::iterator frontIter = scanFront.begin();
           frontIter != scanFront.end();) {
        Instruction* frontInst = *frontIter;

        FwdScan::iterator statusIter = goodDataflows.find(frontInst);
        bool good = goodDataflows.end() != statusIter;
        FwdStatus status = good ? statusIter->second : Invariant;

        if (!good) {
          if (PHINode* phi = dyn_cast<PHINode>(frontInst)) {
            for (int i = 0, iE = phi->getNumIncomingValues(); i < iE; i++) {
              Value* opValue = phi->getIncomingValue(i);
              statusIter = goodDataflows.find(dyn_cast<Instruction>(opValue));
              if (isa<Constant>(opValue) || mInvars.count(opValue) ||
                  (goodDataflows.end() != statusIter)) {
                status = (goodDataflows.end() == statusIter) ? Invariant :
                  (statusIter->second > status ? statusIter->second : status);
                good = true;
              }
            }
          }
          else {
            good = true;
            for (Instruction::op_iterator opIter = frontInst->op_begin(),
                 opIterE = frontInst->op_end(); opIter != opIterE; ++opIter) {
              Instruction* opInst = dyn_cast<Instruction>(*opIter);
              statusIter = goodDataflows.find(opInst);
              if (opInst && !mInvars.count(opInst) &&
                  (goodDataflows.end() == statusIter)) {
                good = false;
                break;
              }
              else if (goodDataflows.end() != statusIter) {
                status = statusIter->second > status ?
                  statusIter->second : status;
              }
            }
          }
        }

        if (good) {
          if (isa<LoadInst>(frontInst)) {
            switch (status) {
              case InstPtr:
                status = InstData; break;
              case Invariant:
              case InstData:
                status = Excluded; break;
              default: break;
            }
          }

          goodDataflows.insert(FwdPair(frontInst, status));

          for (Instruction::use_iterator useIter = frontInst->use_begin(),
               useIterE = frontInst->use_end();
               useIter != useIterE; ++useIter) {
            Instruction* nextInst = cast<Instruction>(*useIter);
            if (!goodDataflows.count(nextInst)) {
              scanFront.insert(nextInst);
            }
          }
        }

        scanFront.erase(frontIter++);
      }
    }

// !!!
errs() << mHead->getName() << "<-name\n";
errs() << goodDataflows.size() << "<-fwd\n";
for (FwdScan::iterator i = goodDataflows.begin(); i != goodDataflows.end(); ++i) {
  errs() << i->second << ":" << *i->first << "\n";
}

    // Backwards
    scanFront.insert(mCtrlFlowVals.begin(), mCtrlFlowVals.end());
    do {
      for (std::set<Instruction*>::iterator frontIter = scanFront.begin();
           frontIter != scanFront.end();) {
        for (Instruction::op_iterator opIter = (*frontIter)->op_begin(),
             opIterE = (*frontIter)->op_end(); opIter != opIterE; ++opIter) {
          if (Instruction* inst = dyn_cast<Instruction>(*opIter)) {
            if (goodDataflows.count(inst)) {
              if (mCtrlFlowVals.insert(inst).second) {
                scanFront.insert(inst);
              }
            }
          }
        }
        const BranchMerge::BranchSet& depsList =
          bmInfo.getBranchesForMerge((*frontIter)->getParent());
        for (BranchMerge::BranchSet::iterator termIter = depsList.begin(),
             termIterE = depsList.end(); termIter != termIterE; ++termIter) {
          Instruction* termInst = *termIter;
          if (goodDataflows.count(termInst)) {
            if (mCtrlFlowVals.insert(termInst).second) {
              scanFront.insert(termInst);
            }
          }
        }

        scanFront.erase(frontIter++);
      }
    } while (!scanFront.empty());

// !!!
errs() << mCtrlFlowVals.size() << "<-bwd\n";
for (BwdScan::iterator i = mCtrlFlowVals.begin();
     i != mCtrlFlowVals.end(); ++i) {
  errs() << "B:" << **i << "\n";
}

    // Find loops which contain purely control-flow logic
    std::set<Instruction*> ctrlFlowLoopsContents;
    for (std::set<Loop*>::iterator loopIter = innerLoops.begin(),
         loopIterE = innerLoops.end(); loopIter != loopIterE; ++loopIter) {
      Loop* loop = *loopIter;
      SmallVector<Instruction*, 32> contents;
      bool ctrlFlowLoop = true;

      for (Loop::block_iterator blkIter = loop->block_begin(),
           blkIterE = loop->block_end();
           ctrlFlowLoop && blkIter != blkIterE; ++blkIter) {
        for (BasicBlock::iterator instIter = (*blkIter)->begin(),
             instIterE = (*blkIter)->end();
             ctrlFlowLoop && instIter != instIterE; ++instIter) {
          if (mCtrlFlowVals.count(&*instIter)) {
            contents.push_back(&*instIter);
          }
          else {
            ctrlFlowLoop = false;
          }
        }
      }

      if (ctrlFlowLoop) {
        mCtrlFlowLoops.insert(loop);
        ctrlFlowLoopsContents.insert(contents.begin(), contents.end());
      }
    }

    // Find which values used to compute new control variants are used outside
    for (std::set<Instruction*>::iterator instIter = mCtrlFlowVals.begin(),
         instIterE = mCtrlFlowVals.end(); instIter != instIterE; ++instIter) {
      Instruction* inst = *instIter;
      bool used = false;

      for (Instruction::use_iterator useIter = inst->use_begin(),
           useIterE = inst->use_end(); useIter != useIterE; ++useIter) {
        if (Instruction* usedInst = dyn_cast<Instruction>(*useIter)) {
          if (!mCtrlFlowVals.count(usedInst) ||
              (ctrlFlowLoopsContents.count(inst) &&
               !ctrlFlowLoopsContents.count(usedInst)) ||
              (isa<TerminatorInst>(usedInst) &&
               outerLoops.count(loopInfo.getLoopFor(usedInst->getParent())))) {
            used = true;
            break;
          }
        }
      }
      if (used) {
        mSpecVals.insert(inst);
      }
    }

// !!!
errs() << mCtrlFlowVals.size() << "<-bwd\n";
for (std::set<Instruction*>::iterator i = mSpecVals.begin(); i != mSpecVals.end(); ++i) {
  errs() << "staticVar:" << **i << "\n";
}

    return false;
  }

  const std::set<CallInst*>& AnalysisPass::getAnnotations() const
  {
    return mAnnotations;
  }

  BasicBlock* AnalysisPass::getHead() const
  {
    return mHead;
  }

  const std::set<BasicBlock*>& AnalysisPass::getBodySet() const
  {
    return mBody;
  }

  const std::list<BasicBlock*>& AnalysisPass::getBodyList() const
  {
    return mBodyList;
  }

  const std::set<Instruction*>& AnalysisPass::getControlFlowValues() const
  {
    return mCtrlFlowVals;
  }

  const std::set<Loop*>& AnalysisPass::getControlFlowLoops() const
  {
    return mCtrlFlowLoops;
  }

  const std::set<Value*>& AnalysisPass::getInvariants() const
  {
    return mInvars;
  }

  const std::set<PHINode*>& AnalysisPass::getControlVariants() const
  {
    return mCtrlVars;
  }

  const std::set<PHINode*>& AnalysisPass::getRuntimeVariants() const
  {
    return mRtVars;
  }

  const std::set<Instruction*>& AnalysisPass::getSpecialiseValues() const
  {
    return mSpecVals;
  }

  void AnalysisPass::getAnalysisUsage(AnalysisUsage& aU) const
  {
    aU.addRequired<BranchMerge>();
    aU.addRequiredTransitive<LoopInfo>();
  }

  char AnalysisPass::ID = 0;
  static RegisterPass<AnalysisPass> X(
    "autojit-analysis",
    "AutoJIT Analysis Pass");
}
