#include <map>
#include <vector>
#include <deque>
#include <cstring>
#include <iostream>
#include <sstream>

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/Constants.h"
#include "llvm/DerivedTypes.h"
#include "llvm/PassManager.h"
#include "llvm/Support/IRBuilder.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/StandardPasses.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetData.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Target/TargetSelect.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ADT/DenseMap.h"
#include "llvm/ADT/SmallVector.h"
using namespace llvm;

enum Status {
  STATUS_CONTINUE = 0,
  STATUS_RETURN   = 1,
  STATUS_UNKNOWN  = 2,
};

typedef struct invars invars_t;
struct ControlVars;
typedef struct rtvars rtvars_t;
typedef struct specvals specvals_t;
typedef struct ctrlinfo ctrlinfo_t;

typedef unsigned int (*CtrlFlowQuant)(
  invars_t*, const ControlVars*, specvals_t*, ctrlinfo_t*);
typedef void (*ConvInvars)(invars_t*, Constant**);
typedef void (*ConvSpecVals)(specvals_t*, Constant**);
typedef void (*ConvCtrlVars)(const ControlVars*, Constant**);

typedef struct {
  CtrlFlowQuant ctrlFlowFunc;
  ConvInvars convInvars;
  ConvSpecVals convSpecVals;
  ConvCtrlVars convCtrlVars;
  char* bitCode;
  unsigned int bitCodeLen;
  void** globalFixups;
  unsigned int numGlobalFixups;
  unsigned int invarsCount;
  unsigned int specValsLen;
  unsigned int specValsCount;
  unsigned int ctrlVarsLen;
  unsigned int ctrlVarsCount;
  unsigned int ctrlInfoLen;
  unsigned int ctrlInfoOffsetVars;
  unsigned int maxCtrlInfo;
} EngineParams;

typedef char (*CompiledCode)(ControlVars*, rtvars_t*);

template <class T>
class MemComparator
{
  size_t mLen;
public:
  MemComparator(size_t len) : mLen(len) {}
  bool operator() (const T* x, const T* y)
  {
    return std::memcmp(x, y, mLen) > 0;
  }
};

const unsigned int MIN_TEMP = 250;
const unsigned int MAX_TEMP = 500;

class AutoJITEngine
{
  struct CacheEntry
  {
    unsigned int temp;
    CompiledCode code;

    CacheEntry()
     : temp(0)
     , code(NULL)
    {}
  };

  struct CompiledInstruction
  {
    typedef std::map<const ControlVars*, CompiledInstruction,
      MemComparator<ControlVars> > Map;

    std::vector<PHINode*> mPhiNodes;

    CompiledInstruction(unsigned int phiCount)
      : mPhiNodes(phiCount)
    {}
  };

  struct PendingInstruction
  {
    const ControlVars* mCtrlVars;
    std::vector<ReturnInst*> mReturns;

    PendingInstruction(const ControlVars* ctrlVars)
      : mCtrlVars(ctrlVars)
    {}

    PendingInstruction(const ControlVars* ctrlVars, ReturnInst* retInst)
      : mCtrlVars(ctrlVars)
    {
      mReturns.push_back(retInst);
    }
  };

  EngineParams engParams;
  LLVMContext& ctx;
  Module* m;
  ExecutionEngine* engine;
  FunctionPassManager* fpm;
  Function* exeTemplate;
  Function* joinMark;
  FunctionType* codeType;
  std::vector<std::vector<ReturnInst*> > mJoinMarks;
  const StructType* ctrlVarsType;
  const StructType* rtVarsType;

  typedef std::map<const ControlVars*, CacheEntry,
    MemComparator<ControlVars> > Cache;
  Cache mCache;

public:
  AutoJITEngine(EngineParams* ep)
    : engParams(*ep)
    , ctx(getGlobalContext())
    , mJoinMarks(engParams.maxCtrlInfo)
    , mCache(MemComparator<ControlVars>(engParams.ctrlVarsLen))
  {
    // Load serialised module
    MemoryBuffer* bitCodeBuffer = MemoryBuffer::getMemBuffer(
      engParams.bitCode, engParams.bitCode+engParams.bitCodeLen-1);
    m = ParseBitcodeFile(bitCodeBuffer, ctx);
    assert(m);
    delete bitCodeBuffer;

    // Create JIT engine
    InitializeNativeTarget();
    engine = EngineBuilder(m).setOptLevel(CodeGenOpt::None).create();
    assert(engine);

    // Create optimisation pipeline
    fpm = new FunctionPassManager(m);
    fpm->add(new TargetData(*engine->getTargetData()));
    fpm->add(createCFGSimplificationPass());
    fpm->add(createInstructionCombiningPass());
    fpm->add(createReassociatePass());
    fpm->add(createGVNPass());
    fpm->add(createSCCPPass());
    fpm->add(createPromoteMemoryToRegisterPass());
    fpm->add(createDeadStoreEliminationPass());
    fpm->add(createInstructionCombiningPass());
    //fpm->add(createAggressiveDCEPass());
    /*fpm->add(createLoopRotatePass());
    fpm->add(createLICMPass()); 
    fpm->add(createInstructionCombiningPass());
    fpm->add(createIndVarSimplifyPass());
    fpm->add(createLoopUnrollPass());
    fpm->add(createInstructionCombiningPass());
    fpm->add(createCFGSimplificationPass());*/
    fpm->doInitialization();

    // Setup global mappings
    for (int i=0; i<engParams.numGlobalFixups; i++) {
      std::stringstream s;
      s << "__g" << i;
      std::string name = s.str();
      GlobalValue* global =
        dyn_cast_or_null<GlobalValue>(m->getFunction(name));
      if (!global) {
        global = dyn_cast_or_null<GlobalValue>(m->getNamedGlobal(name));
      }
      if (!global) {
        global = dyn_cast_or_null<GlobalValue>(m->getNamedAlias(name));
      }
      assert(global);
      engine->addGlobalMapping(global, engParams.globalFixups[i]);
    }

    // Aquire function handles
    exeTemplate = m->getFunction("template");
    assert(exeTemplate);
    joinMark = m->getFunction("joinmark");
    assert(joinMark);

    // Save join marks
    for (inst_iterator instIter = inst_begin(exeTemplate),
         instIterE = inst_end(exeTemplate);
         instIter != instIterE; ++instIter) {
      if (ReturnInst* retInst = dyn_cast<ReturnInst>(&*instIter)) {
        if (MDNode* md = retInst->getMetadata("autojit")) {
          unsigned int id =
            cast<ConstantInt>(md->getOperand(0))->getZExtValue();
          mJoinMarks[id].push_back(retInst);
        }
      }
    }

    // Create structure types
    const FunctionType* exeType = exeTemplate->getFunctionType();
    unsigned int index = engParams.invarsCount+engParams.specValsCount;
    std::vector<const Type*> ctrlVarsTypes;
    for (unsigned int iE = index+engParams.ctrlVarsCount; index < iE; index++) {
      ctrlVarsTypes.push_back(exeType->getParamType(index));
    }
    ctrlVarsType = StructType::get(ctx, ctrlVarsTypes);
    std::vector<const Type*> rtVarsTypes;
    for (unsigned int iE = exeType->getNumParams(); index < iE; index++) {
      rtVarsTypes.push_back(exeType->getParamType(index));
    }
    rtVarsType = StructType::get(ctx, rtVarsTypes);

    // Create code function type
    std::vector<const Type*> codeTypes;
    codeTypes.push_back(PointerType::getUnqual(ctrlVarsType));
    codeTypes.push_back(PointerType::getUnqual(rtVarsType));
    codeType = FunctionType::get(
      Type::getInt8Ty(ctx), codeTypes, false);
  }

  void capInstruction(
    ReturnInst* retInst,
    Argument* ctrlVarsArg,
    Argument* rtVarsArg)
  {
    IRBuilder<> builder(retInst->getParent());
    if (CallInst* callInst = dyn_cast<CallInst>(retInst->getReturnValue())) {
      if (callInst->getCalledFunction() == joinMark) {
        for (unsigned int i=0; i<ctrlVarsType->getNumElements(); i++) {
          builder.CreateStore(callInst->getOperand(1+i),
            builder.CreateStructGEP(ctrlVarsArg, i));
        }
        for (unsigned int i=0; i<rtVarsType->getNumElements(); i++) {
          builder.CreateStore(
            callInst->getOperand(1+ctrlVarsType->getNumElements()+i),
            builder.CreateStructGEP(rtVarsArg, i));
        }
        builder.CreateRet(ConstantInt::get(Type::getInt8Ty(ctx), 0));
        retInst->eraseFromParent();
        callInst->eraseFromParent();
        return;
      }
    }

    builder.CreateStore(retInst->getReturnValue(),
      builder.CreateBitCast(rtVarsArg,
        PointerType::getUnqual(retInst->getReturnValue()->getType())));
    builder.CreateRet(ConstantInt::get(Type::getInt8Ty(ctx), 1));
    retInst->eraseFromParent();
  }

  unsigned int insertInstruction(
    invars_t* invars,
    const ControlVars* ctrlVars,
    ctrlinfo_t* ctrlInfo,
    PHINode** phiNodes,
    DenseMap<const Value*, Value*>& valueMap,
    Function* code)
  {
    char specVals[engParams.specValsLen];
    unsigned int rets = engParams.ctrlFlowFunc(
      invars, ctrlVars, (specvals_t*)specVals, ctrlInfo);

    Constant* invarsArray[engParams.invarsCount];
    Constant* specValsArray[engParams.specValsCount];
    Constant* ctrlVarsArray[engParams.ctrlVarsCount];
    engParams.convInvars((invars_t*)invars, invarsArray);
    engParams.convSpecVals((specvals_t*)specVals, specValsArray);
    engParams.convCtrlVars(ctrlVars, ctrlVarsArray);

    iplist<Argument>::iterator oldArgs = exeTemplate->getArgumentList().begin();
    SmallVector<ReturnInst*, 16> returns;

    BasicBlock* bridgeBlock = BasicBlock::Create(ctx, "bridge", code);
    IRBuilder<> builder(bridgeBlock);
    for (unsigned int i=0; i<engParams.invarsCount; i++) {
      if (invarsArray[i]->getType() != oldArgs->getType()) {
        invarsArray[i] = ConstantExpr::getBitCast(
          invarsArray[i], oldArgs->getType());
      }
      valueMap[&*oldArgs++] = invarsArray[i];
    }
    for (unsigned int i=0; i<engParams.specValsCount; i++) {
assert(oldArgs->getType() == specValsArray[i]->getType());
      valueMap[&*oldArgs++] = specValsArray[i];
    }
    for (unsigned int i=0; i<ctrlVarsType->getNumElements(); i++) {
      valueMap[&*oldArgs++] = ctrlVarsArray[i];
    }
    for (unsigned int i=0; i<rtVarsType->getNumElements(); i++) {
      valueMap[&*oldArgs++] = phiNodes[i] =
        builder.CreatePHI(rtVarsType->getElementType(i));
    }
    CloneFunctionInto(code, exeTemplate, valueMap, returns);
    builder.CreateBr(cast<BasicBlock>(valueMap[&exeTemplate->getEntryBlock()]));
    return rets;
  }

  void joinInstructions(
    ReturnInst* retInst,
    PHINode** phiNodes)
  {
    CallInst* callInst = dyn_cast<CallInst>(retInst->getReturnValue());
    if (!callInst || joinMark != callInst->getOperand(0)) {
      // Skip over bad join points that will be later removed by specialisation
      return;
    }
    for (unsigned int i=0; i<rtVarsType->getNumElements(); i++) {
      phiNodes[i]->addIncoming(
        callInst->getOperand(1+ctrlVarsType->getNumElements()+i),
        callInst->getParent());
    }
    IRBuilder<> builder(callInst->getParent());
    builder.CreateBr(phiNodes[0]->getParent());
    retInst->eraseFromParent();
    callInst->eraseFromParent();
  }

  CacheEntry& getCacheEntry(const ControlVars*& ctrlVars)
  {
    Cache::iterator entryIter = mCache.find(ctrlVars);
    if (mCache.end() == entryIter) {
      void* key = new char[engParams.ctrlVarsLen];
      std::memcpy(key, ctrlVars, engParams.ctrlVarsLen);
      ctrlVars = (ControlVars*)key;
      return mCache[ctrlVars];
    }
    else {
      ctrlVars = entryIter->first;
      return entryIter->second;
    }
  }

  CompiledCode compileCode(
    invars_t* invars,
    const ControlVars* ctrlVars,
    rtvars_t* rtVars)
  {
    Function* code = Function::Create(
      codeType, GlobalValue::PrivateLinkage, "code", m);
    iplist<Argument>::iterator args = code->getArgumentList().begin();
    Argument* ctrlVarsArg = &*args++;
    Argument* rtVarsArg = &*args++;

    BasicBlock* entryBlock = BasicBlock::Create(ctx, "entry", code);
    IRBuilder<> builder(entryBlock);

    std::vector<Value*> rtValues;
    for (unsigned int i=0; i<ctrlVarsType->getNumElements(); i++) {
      rtValues.push_back(UndefValue::get(ctrlVarsType->getElementType(i)));
    }
    for (unsigned int i=0; i<rtVarsType->getNumElements(); i++) {
      rtValues.push_back(builder.CreateLoad(builder.CreateStructGEP(
        rtVarsArg, i)));
    }
    ReturnInst* retInst = builder.CreateRet(builder.CreateCall(
      joinMark, rtValues.begin(), rtValues.end()));

    // Assemble function
    CompiledInstruction::Map compiledInsts(
      MemComparator<ControlVars>(engParams.ctrlVarsLen));
    std::deque<PendingInstruction> pendingInsts;
    pendingInsts.push_front(PendingInstruction(ctrlVars, retInst));
    do {
      PendingInstruction& pending = pendingInsts.front();

      // Append specialised instruction template to function
      CompiledInstruction compiled(rtVarsType->getNumElements());
      DenseMap<const Value*, Value*> valueMap;
      char ctrlInfo[engParams.ctrlInfoLen*engParams.maxCtrlInfo];
      unsigned int rets = insertInstruction(
        invars, pending.mCtrlVars,
        (ctrlinfo_t*)ctrlInfo, &compiled.mPhiNodes[0],
        valueMap, code);
      compiledInsts.insert(CompiledInstruction::Map::value_type(
        pending.mCtrlVars, compiled));

      // Join instruction to pending predecessors
      for (std::vector<ReturnInst*>::iterator retIter = pending.mReturns.begin(),
           retIterE = pending.mReturns.end(); retIter != retIterE; ++retIter) {
        joinInstructions(*retIter, &compiled.mPhiNodes[0]);
      }

      // Join instruction to successors or cap it as necessary
      for (unsigned int i=0; i<rets; i++) {
        std::vector<ReturnInst*>& joinMarks = mJoinMarks[i];

        if (STATUS_CONTINUE == ctrlInfo[i*engParams.ctrlInfoLen]) {
          const ControlVars* nextCtrlVars = (ControlVars*)&ctrlInfo[
            i*engParams.ctrlInfoLen + engParams.ctrlInfoOffsetVars];
          CompiledInstruction::Map::iterator compIter =
            compiledInsts.find(nextCtrlVars);
          if (compiledInsts.end() == compIter) {
            CacheEntry& entry = getCacheEntry(nextCtrlVars);
            if (entry.temp >= MIN_TEMP) {
              PendingInstruction nextPending(nextCtrlVars);
              for (std::vector<ReturnInst*>::iterator retIter = joinMarks.begin(),
                   retIterE = joinMarks.end(); retIter != retIterE; ++retIter) {
                nextPending.mReturns.push_back(
                  cast<ReturnInst>(valueMap[*retIter]));
              }
              pendingInsts.push_back(nextPending);
            }
            else {
              for (std::vector<ReturnInst*>::iterator retIter = joinMarks.begin(),
                   retIterE = joinMarks.end(); retIter != retIterE; ++retIter) {
                capInstruction(
                  cast<ReturnInst>(valueMap[*retIter]), ctrlVarsArg, rtVarsArg);
              }
            }
          }
          else {
            CompiledInstruction& compInst = compIter->second;
            for (std::vector<ReturnInst*>::iterator retIter = joinMarks.begin(),
                 retIterE = joinMarks.end(); retIter != retIterE; ++retIter) {
              joinInstructions(
                cast<ReturnInst>(valueMap[*retIter]), &compInst.mPhiNodes[0]);
            }
          }
        }
        else {
          for (std::vector<ReturnInst*>::iterator retIter = joinMarks.begin(),
               retIterE = joinMarks.end(); retIter != retIterE; ++retIter) {
            capInstruction(
              cast<ReturnInst>(valueMap[*retIter]), ctrlVarsArg, rtVarsArg);
          }
        }
      }

      pendingInsts.pop_front();
    } while (!pendingInsts.empty());

//errs() << "optimise\n";
    fpm->run(*code);
//errs() << "generate\n";

//errs() << *code;
    CompiledCode ccode = (CompiledCode)engine->getPointerToFunction(code);
//errs() << "ready\n";
    //code->eraseFromParent();
    return ccode;
  }

  char process(
    invars_t* invars,
    ControlVars* ctrlVars,
    rtvars_t* rtVars)
  {
    const ControlVars* constCtrlVars = ctrlVars;
    CacheEntry& entry = getCacheEntry(constCtrlVars);
    if (entry.temp < MAX_TEMP) {
      entry.temp++;
      return 2;
    }
    else {
      if (entry.temp == MAX_TEMP) {
        entry.temp++;
        entry.code = compileCode(invars, ctrlVars, rtVars);
      }
      return (*entry.code)(ctrlVars, rtVars);
    }
  }
};

static __thread LLVMContext* context = NULL;
static __thread AutoJITEngine* engine = NULL;

extern "C" void* autojit_get_integer(int bits, long value)
{
  return ConstantInt::get(IntegerType::get(*context, bits), value);
}

extern "C" void* autojit_get_pointer(void* value)
{
  return ConstantExpr::getIntToPtr(ConstantInt::get(IntegerType::get(
    *context, 8*sizeof(value)), (uint64_t)value), Type::getInt8PtrTy(*context));
}

extern "C" char autojit_external_engine(
  EngineParams* engParams,
  invars_t* invars,
  ControlVars* ctrlVars,
  rtvars_t* rtVars)
{
  if (!engine) {
    std::cout << "Engine created." << std::endl;

    context = &getGlobalContext();
    engine = new AutoJITEngine(engParams);
  }

  return engine->process(invars, ctrlVars, rtVars);
}
