//===- BranchMerge.cpp - Branch/Merge Correspondence ------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "branch-merge"

#include <map>
#include <set>
#include <vector>

#include "BranchMerge.h"

#include "llvm/Support/raw_ostream.h"
using namespace llvm;

namespace autojit {
  struct Splinter {
    TerminatorInst* termInst;
    unsigned int branch;
  };

  bool operator<(const Splinter& lhs, const Splinter& rhs) {
    return (lhs.termInst == rhs.termInst) ?
      lhs.branch < rhs.branch : lhs.termInst < rhs.termInst;
  }

  struct BlockInfo {
    typedef std::set<Splinter>::iterator iterator;

    std::set<Splinter> splinters;
    std::set<TerminatorInst*> merged;

    iterator begin() {return splinters.begin();}
    iterator end() {return splinters.end();}
  };

  bool BranchMerge::runOnFunction(Function& f)
  {
    typedef std::map<BasicBlock*,BlockInfo> BlockMap;
    typedef std::set<BasicBlock*> Front;

    BlockMap blockMap;
    Front front;
    front.insert(&f.getEntryBlock());
    do {
      for (Front::iterator frontIter = front.begin();
           frontIter != front.end();) {
        BasicBlock* block = *frontIter;
        BlockInfo& blockInfo = blockMap[block];
        TerminatorInst* termInst = block->getTerminator();

        // Test for complete sets of splinters
        bool mergedChanged = false;
        for (BlockInfo::iterator splIter = blockInfo.splinters.begin();
             splIter != blockInfo.splinters.end();) {
          TerminatorInst* upstreamTermInst = splIter->termInst;

          // Count how may splinters for the current upstream terminator
          unsigned int count = 0;
          for (; splIter != blockInfo.splinters.end() &&
            splIter->termInst == upstreamTermInst; ++splIter) {
            count++;
          }

          // Record the merge if the splinter set is complete#
          if (upstreamTermInst->getNumSuccessors() == count) {
            if (blockInfo.merged.insert(upstreamTermInst).second) {
              mBranchSets[block].insert(upstreamTermInst);
              mMergeSets[upstreamTermInst].insert(block);
              mergedChanged = true;
            }
          }
        }

        // Update successors
        for (unsigned int i = 0, iE = termInst->getNumSuccessors();
             i < iE; i++) {
          BasicBlock* succ = termInst->getSuccessor(i);
          BlockInfo& succInfo = blockMap[succ];
          bool splintersChanged = false;

          // Purge splinters which have just been merged in the predecessor
          if (mergedChanged) {
            for (BlockInfo::iterator splIter = succInfo.splinters.begin(),
                 splIterE = succInfo.splinters.end(); splIter != splIterE;) {
              if (blockInfo.merged.count(splIter->termInst)) {
                 succInfo.splinters.erase(splIter++);
                splintersChanged = true;
              }
              else {
                ++splIter;
              }
            }
          }

          // Propagate unmerged splinters from the predecessor
          for (BlockInfo::iterator splIter = blockInfo.splinters.begin(),
               splIterE = blockInfo.splinters.end();
               splIter != splIterE; ++splIter) {
            if (!blockInfo.merged.count(splIter->termInst)) {
              splintersChanged |= succInfo.splinters.insert(*splIter).second;
            }
          }

          // Record the branch from the predecessor if it is multi-way
          if (iE > 1) {
            Splinter splint;
            splint.termInst = termInst;
            splint.branch = i;
            splintersChanged |= succInfo.splinters.insert(splint).second;
          }
          else if (iE == 1) {
            // Propagate front along linear series of blocks
            splintersChanged = true;
          }

          // Mark blocks with an altered splinter set for further processing
          if (splintersChanged) {
            front.insert(succ);
          }
        }

        front.erase(frontIter++);
      }
    } while (!front.empty());

    return false;
  }

  const BranchMerge::BranchSet& BranchMerge::getBranchesForMerge(
    BasicBlock* block)
  {
    return mBranchSets[block];
  }

  const BranchMerge::MergeSet& BranchMerge::getMergesForBranch(
    TerminatorInst* termInst)
  {
    return mMergeSets[termInst];
  }

  const BranchMerge::MergeSet& BranchMerge::getUnmerged()
  {
    return mUnmerged;
  }

  void BranchMerge::getAnalysisUsage(AnalysisUsage& aU) const
  {
  }

  char BranchMerge::ID = 0;
  static RegisterPass<BranchMerge> X(
    "branch-merge",
    "Branch/Merge Correspondence");
}
