//===- TransformPass.cpp - AutoJIT Transformation Pass ----------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "autojit"

#include <deque>
#include <list>
#include <sstream>
#include <cstdarg>

#include "AnalysisPass.h"
#include "TransformPass.h"

#include "llvm/Type.h"
#include "llvm/Function.h"
#include "llvm/PassManager.h"
#include "llvm/Constants.h"
#include "llvm/Metadata.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Target/TargetData.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Local.h"
#include "llvm/Transforms/Utils/ValueMapper.h"
#include "llvm/Support/IRBuilder.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Bitcode/BitstreamWriter.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/ADT/StringRef.h"
#include "llvm/ADT/ValueMap.h"
using namespace llvm;

namespace autojit {
  struct Tranche
  {
    BasicBlock* head;
    BasicBlock* prehead;
    DenseMap<const Value*,Value*> valueMap;
    Tranche* parent;
    std::vector<Tranche*> children;

    std::vector<BasicBlock*> attachPoints;
    Tranche* nextPeer;
    unsigned int infoSlot;

    void setValueMap(const Value* oldValue, Value* newValue)
    {
      valueMap[oldValue] = newValue;
      for (std::vector<Tranche*>::iterator tIter = children.begin(),
           tIterE = children.end(); tIter != tIterE; ++tIter) {
        (*tIter)->setValueMap(oldValue, newValue);
      }
    }
  };

  Value* buildGEP(
    IRBuilder<>& builder, Value* val, unsigned int numIndicies, ...)
  {
    va_list args;
    va_start(args, numIndicies);

    std::vector<Value*> indicies;
    for (unsigned int i = 0; i < numIndicies; i++) {
      indicies.push_back(ConstantInt::get(Type::getInt32Ty(val->getContext()),
        va_arg(args, unsigned int)));
    }
    return builder.CreateGEP(val, indicies.begin(), indicies.end());
  }

  void fixupGlobals(
    Value* value,
    DenseMap<const Value*, Value*>& valueMap,
    std::vector<Constant*>& globalFixups,
    Module* newModule)
  {
    if (Function* funcSrc = dyn_cast<Function>(value)) {
      if (valueMap.find(funcSrc) == valueMap.end()) {
        std::stringstream nameBuilder;
        nameBuilder << "__g" << globalFixups.size();
        Function* funcDecl = Function::Create(
          funcSrc->getFunctionType(),
          GlobalValue::ExternalLinkage,
          nameBuilder.str(),
          newModule);
        funcDecl->copyAttributesFrom(funcSrc);
        valueMap[funcSrc] = funcDecl;
        globalFixups.push_back(cast<Constant>(funcSrc));
      }
    }
    else if (GlobalVariable* gvarSrc = dyn_cast<GlobalVariable>(value)) {
      if (valueMap.find(funcSrc) == valueMap.end()) {
        std::stringstream nameBuilder;
        nameBuilder << "__g" << globalFixups.size();
        GlobalVariable* gvarDecl = new GlobalVariable(
          gvarSrc->getType()->getElementType(),
          gvarSrc->isConstant(),
          GlobalValue::ExternalLinkage,
          NULL,
          nameBuilder.str(),
          newModule);
        gvarDecl->copyAttributesFrom(gvarSrc);
        valueMap[gvarSrc] = gvarDecl;
        globalFixups.push_back(cast<Constant>(gvarSrc));
      }
    }
    else if (const ConstantExpr* user = dyn_cast<ConstantExpr>(value)) {
     for (ConstantExpr::const_op_iterator opIter = user->op_begin(),
           opIterE = user->op_end(); opIter != opIterE; ++opIter) {
        fixupGlobals(opIter->get(), valueMap, globalFixups, newModule);
      }
    }
    else {
      assert(!isa<GlobalValue>(*value));
    }
  }

  class FuncState {
    Function& srcFunc;
    Pass& thisPass;
    AnalysisPass& analysis;
    TargetData& targetData;
    LLVMContext& ctx;
    Module* mod;

    const StructType* invarsType;
    const StructType* ctrlVarsType;
    const StructType* specValsType;
    const StructType* ctrlInfoType;
    const UnionType* rtStateType;
    Function* controlFlowQuant;
    unsigned int numCtrlInfoSlots;
    GlobalVariable* templateBitCode;
    unsigned int templateBitCodeLen;
    GlobalVariable* templateGlobals;
    unsigned int templateGlobalsCount;
    Function* getInt;
    Function* getPtr;
    Function* convertInvars;
    Function* convertSpecVals;
    Function* convertCtrlVars;
    GlobalVariable* engineInfo;
    Function* externalEngine;

    Function* generateConverter(const StructType* structType, StringRef name)
    {
      // Create function
      std::vector<const Type*> argTypes;
      argTypes.push_back(PointerType::getUnqual(structType));
      argTypes.push_back(PointerType::getUnqual(Type::getInt8PtrTy(ctx)));
      FunctionType* funcType =
        FunctionType::get(Type::getVoidTy(ctx), argTypes, false);
      Function* func = Function::Create(
        funcType, GlobalValue::ExternalLinkage, prefix(name), mod);

      // Populate function body
      iplist<Argument>::iterator args = func->getArgumentList().begin();
      Argument* structArg = &*args++;
      Argument* arrayArg = &*args++;
      BasicBlock* entry = BasicBlock::Create(ctx, "entry", func);
      IRBuilder<> builder(entry);
      for (unsigned int i=0, iE = structType->getNumElements(); i<iE; i++) {
        Value* val = builder.CreateLoad(
        builder.CreateStructGEP(structArg, i));
        Value* converted;
        const Type* valType = val->getType();
        if (const IntegerType* intType = dyn_cast<IntegerType>(valType)) {
          converted = builder.CreateCall2(getGetInt(),
            getConst32(intType->getBitWidth()),
            builder.CreateZExt(val, Type::getInt64Ty(ctx)));
        }
        else if (isa<PointerType>(valType)) {
          converted = builder.CreateCall(getGetPtr(),
            builder.CreateBitCast(val, Type::getInt8PtrTy(ctx)));
        }
        else {
          // Unsupported type
          assert(0);
        }
        builder.CreateStore(converted,
          builder.CreateConstGEP1_32(arrayArg, i));
      }
      builder.CreateRetVoid();
      return func;
    }

public:
    const IntegerType* const tyi8;
    const IntegerType* const tyi32;

    FuncState(Function& sf, Pass& pass, AnalysisPass& ap)
      : srcFunc(sf)
      , thisPass(pass)
      , analysis(ap)
      , targetData(pass.getAnalysis<TargetData>())
      , ctx(sf.getContext())
      , tyi8(Type::getInt8Ty(ctx))
      , tyi32(Type::getInt32Ty(ctx))
    {
      mod = srcFunc.getParent();
      invarsType = NULL;
      ctrlVarsType = NULL;
      specValsType = NULL;
      ctrlInfoType = NULL;
      rtStateType = NULL;
      controlFlowQuant = NULL;
      templateBitCode = NULL;
      templateGlobals = NULL;
      getInt = NULL;
      getPtr = NULL;
      convertInvars = NULL;
      convertSpecVals = NULL;
      convertCtrlVars = NULL;
      engineInfo = NULL;
      externalEngine = NULL;
    }

    std::string prefix(StringRef name)
    {
      return (srcFunc.getName()+"__autojit__"+name).str();
    }

    inline ConstantInt* getConst8(unsigned char value)
    {
      return ConstantInt::get(tyi8, value);
    }

    inline ConstantInt* getConst32(unsigned int value)
    {
      return ConstantInt::get(tyi32, value);
    }

    const StructType* getInvarsType()
    {
      if (invarsType) {
        return invarsType;
      }
      const std::set<Value*>& invariants = analysis.getInvariants();
      std::vector<const Type*> invarTypes;
      for (std::set<Value*>::iterator invarIter = invariants.begin(),
           invarIterE = invariants.end(); invarIter != invarIterE;
           ++invarIter) {
        invarTypes.push_back((*invarIter)->getType());
      }
      invarsType = StructType::get(ctx, invarTypes);
      mod->addTypeName(prefix("invars"), invarsType);
      return invarsType;
    }

    const StructType* getCtrlVarsType()
    {
      if (ctrlVarsType) {
        return ctrlVarsType;
      }
      const std::set<PHINode*>& controlVariants =
        analysis.getControlVariants();
      std::vector<const Type*> ctrlVarTypes;
      for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
           phiIterE = controlVariants.end(); phiIter != phiIterE; ++phiIter) {
        ctrlVarTypes.push_back((*phiIter)->getType());
      }
      ctrlVarsType = StructType::get(ctx, ctrlVarTypes);
      mod->addTypeName(prefix("ctrlvars"), ctrlVarsType);
      return ctrlVarsType;
    }

    const StructType* getSpecValsType()
    {
      if (specValsType) {
        return specValsType;
      }
      const std::set<Instruction*>& specVals = analysis.getSpecialiseValues();
      std::vector<const Type*> specValsTypes;
      for (std::set<Instruction*>::iterator valIter = specVals.begin(),
           valIterE = specVals.end(); valIter != valIterE; ++valIter) {
        specValsTypes.push_back((*valIter)->getType());
      }
      specValsType = StructType::get(ctx, specValsTypes);
      mod->addTypeName(prefix("specvals"), specValsType);
      return specValsType;
    }

    const StructType* getCtrlInfoType()
    {
      if (ctrlInfoType) {
        return ctrlInfoType;
      }
      ctrlInfoType = StructType::get(
        ctx, tyi8, getCtrlVarsType(), NULL);
      mod->addTypeName(prefix("ctrlinfo"), ctrlInfoType);
      return ctrlInfoType;
    }

    const UnionType* getRtStateType()
    {
      if (rtStateType) {
        return rtStateType;
      }
      const std::set<PHINode*>& runtimeVariants = analysis.getRuntimeVariants();
      std::vector<const Type*> rtVarsTypes;
      for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
           phiIterE = runtimeVariants.end(); phiIter != phiIterE; ++phiIter) {
        rtVarsTypes.push_back((*phiIter)->getType());
      }
      rtStateType = UnionType::get(
        StructType::get(ctx, rtVarsTypes), srcFunc.getReturnType(), NULL);
      mod->addTypeName(prefix("rtstate"), rtStateType);
      return rtStateType;
    }

    Function* getControlFlowQuant()
    {
      if (controlFlowQuant) {
        return controlFlowQuant;
      }

      // Create control-flow function
      std::vector<const Type*> argTypes;
      argTypes.push_back(PointerType::getUnqual(getInvarsType()));
      argTypes.push_back(PointerType::getUnqual(getCtrlVarsType()));
      argTypes.push_back(PointerType::getUnqual(getSpecValsType()));
      argTypes.push_back(PointerType::getUnqual(getCtrlInfoType()));
      FunctionType* funcType =
        FunctionType::get(tyi32, argTypes, false);
      controlFlowQuant = Function::Create(
        funcType, GlobalValue::ExternalLinkage, prefix("cfq"), mod);


      // Create root tranche and entry block
      DenseMap<BasicBlock*,Tranche*> blockTranche;
      Tranche* rootTranche = new Tranche();
      rootTranche->head = analysis.getHead();
      BasicBlock* entryBlock = BasicBlock::Create(
        ctx, "entry", controlFlowQuant);
      rootTranche->valueMap[entryBlock] = entryBlock;
      rootTranche->prehead = entryBlock;
      blockTranche[entryBlock] = rootTranche;
      IRBuilder<> builder(entryBlock);

      // Emit loads for invariants
      iplist<Argument>::iterator args =
        controlFlowQuant->getArgumentList().begin();
      Argument* invarArg = &*args++;
      rootTranche->valueMap[invarArg] = invarArg;
      int invarOff = 0;
      const std::set<Value*>& invariants = analysis.getInvariants();
      for (std::set<Value*>::iterator invarIter = invariants.begin(),
           invarIterE = invariants.end(); invarIter != invarIterE; ++invarIter) {
        Value* gep = builder.CreateStructGEP(invarArg, invarOff++);
        rootTranche->valueMap[gep] = gep;
        rootTranche->valueMap[*invarIter] = builder.CreateLoad(gep);
      }

      // Emit loads for control variants
      Argument* ctrlVarArg = &*args++;
      DenseMap<Value*,Value*> phiMap;
      rootTranche->valueMap[ctrlVarArg] = ctrlVarArg;
      int ctrlVarOff = 0;
      const std::set<PHINode*>& controlVariants = analysis.getControlVariants();
      for (std::set<PHINode*>::iterator ctrlVarIter = controlVariants.begin(),
           ctrlVarIterE = controlVariants.end();
           ctrlVarIter != ctrlVarIterE; ++ctrlVarIter) {
        Value* gep = builder.CreateStructGEP(ctrlVarArg, ctrlVarOff++);
        rootTranche->valueMap[gep] = gep;
        rootTranche->valueMap[*ctrlVarIter] = builder.CreateLoad(gep);
      }

      // Append dummy terminator and setup other arguments
      builder.CreateRetVoid();
      Argument* specValsArg = &*args++;
      rootTranche->valueMap[specValsArg] = specValsArg;
      Argument* ctrlInfoArg = &*args++;
      rootTranche->valueMap[ctrlInfoArg] = ctrlInfoArg;

      // Generate the control-flow quantifier
      unsigned int maxSlot = 0;
      const std::set<Instruction*>& specVals = analysis.getSpecialiseValues();
      std::deque<Tranche*> tranches;
      tranches.push_front(rootTranche);
      while (!tranches.empty()) {
        std::set<BasicBlock*> trancheBody, trancheFront;
        std::set<BasicBlock*> trancheJoins, trancheTails;
        Tranche* tranche = tranches.front();
        tranches.pop_front();
        trancheFront.insert(tranche->head);
        maxSlot = tranche->infoSlot > maxSlot ? tranche->infoSlot : maxSlot;

        // Find the extent of the next tranche
        do {
          for (std::set<BasicBlock*>::iterator frontIter = trancheFront.begin(),
               frontIterE = trancheFront.end(); frontIter != frontIterE;) {
            BasicBlock* block = *frontIter;
            TerminatorInst* termInst = block->getTerminator();
            trancheBody.insert(block);

            Value* condition = NULL;
            if (BranchInst* brInst = dyn_cast<BranchInst>(termInst)) {
              condition = brInst->isConditional() ?
                brInst->getCondition() : NULL;
            }
            else if (SwitchInst* swInst = dyn_cast<SwitchInst>(termInst)) {
              condition = swInst->getCondition();
            }

            if (condition && isa<Instruction>(condition) &&
                !analysis.getControlFlowValues().
                  count(cast<Instruction>(condition))) {
              trancheJoins.insert(block);
              Tranche* nextPeer = NULL;
              for (Tranche* t = tranche; t; t = t->parent) {
                if  (t->nextPeer) {
                  nextPeer = t->nextPeer;
                  break;
                }
              }
              for (succ_iterator succIter = succ_begin(block),
                   succIterE = succ_end(block);
                   succIter != succIterE; ++succIter) {
                BasicBlock* succ = *succIter;
                Tranche* newTranche = new Tranche();
                newTranche->head = succ;
                newTranche->prehead = block;
                newTranche->parent = tranche;
                newTranche->nextPeer = nextPeer;
                newTranche->infoSlot = tranche->infoSlot;
                nextPeer = newTranche;
                tranche->children.push_back(newTranche);
                tranches.push_front(newTranche);
              }
            }
            else {
              if (!termInst->getNumSuccessors()) {
                trancheTails.insert(block);
              }
              else {
                for (succ_iterator succIter = succ_begin(block),
                     succIterE = succ_end(block);
                     succIter != succIterE; ++succIter) {
                  BasicBlock* succ = *succIter;
                  if (analysis.getHead() == succ) {
                    trancheTails.insert(block);
                  }
                  else if (!trancheBody.count(succ)) {
                    trancheFront.insert(succ);
                  }
                }
              }
            }

            trancheFront.erase(frontIter++);
          }
        } while (!trancheFront.empty());

        // Copy current tranche into destination function
        const std::list<BasicBlock*>& bodyList = analysis.getBodyList();
        for (std::list<BasicBlock*>::const_iterator blkIter = bodyList.begin(),
             blkIterE = bodyList.end(); blkIter != blkIterE; ++blkIter) {
          BasicBlock* oldBlk = *blkIter;
          if (!trancheBody.count(oldBlk)) {
            continue;
          }
          BasicBlock* newBlk = BasicBlock::Create(
            ctx, oldBlk->hasName()?oldBlk->getName():"", controlFlowQuant);
          tranche->valueMap[oldBlk] = newBlk;
          blockTranche[newBlk] = tranche;
          for (BasicBlock::iterator instIter = oldBlk->begin(),
               instIterE = oldBlk->end(); instIter != instIterE; ++instIter) {
            Instruction* oldInst = &*instIter;
            if (!controlVariants.count(dyn_cast<PHINode>(oldInst))) {
              if (analysis.getControlFlowValues().count(oldInst) ||
                  isa<TerminatorInst>(oldInst)) {

                // Copy instruction
                Instruction* newInst = oldInst->clone();
                if (oldInst->hasName()) {
                  newInst->setName(oldInst->getName());
                }
                newBlk->getInstList().push_back(newInst);
                tranche->valueMap[oldInst] = newInst;

                // Remove broken values if instruction is a phi node
                if (PHINode* phi = dyn_cast<PHINode>(newInst)) {
                  for (unsigned int i=0; i<phi->getNumIncomingValues(); i++) {
                    BasicBlock* pred = phi->getIncomingBlock(i);
                    if ((!trancheBody.count(pred) || trancheJoins.count(pred)) &&
                        (tranche->prehead != pred || tranche->head != oldBlk)) {
                      phi->removeIncomingValue(i--);
                    }
                  }
                }
              }
              else {
                tranche->valueMap[oldInst] = UndefValue::get(oldInst->getType());
              }
            }
          }

          // Save values for later use in specialisation
          for (BasicBlock::iterator instIter = oldBlk->begin(),
               instIterE = oldBlk->end(); instIter != instIterE; ++instIter) {
            Instruction* oldInst = &*instIter;
            if (specVals.count(oldInst)) {
              unsigned int idx =
                std::distance(specVals.begin(), specVals.find(oldInst));
              IRBuilder<> builder(newBlk, --newBlk->end());
              Value* gep = builder.CreateStructGEP(specValsArg, idx);
              tranche->valueMap[gep] = gep;
              builder.CreateStore(oldInst, gep);
            }
          }
        }

        // Add bridge blocks to tails
        for (std::set<BasicBlock*>::iterator tailIter =
             trancheTails.begin(), tailIterE = trancheTails.end();
             tailIter != tailIterE; ++tailIter) {
          BasicBlock* oldTail = *tailIter;
          BasicBlock* tailBlock = cast<BasicBlock>(tranche->valueMap[oldTail]);
          TerminatorInst* termInst = tailBlock->getTerminator();
          BasicBlock* bridgeBlock = BasicBlock::Create(
            ctx, "bridge", controlFlowQuant);
          IRBuilder<> builder(bridgeBlock);
          Value* gep = builder.CreateConstGEP2_32(ctrlInfoArg, tranche->infoSlot, 0);
          tranche->valueMap[gep] = gep;
          builder.CreateStore(getConst8(
            isa<ReturnInst>(termInst) ? 1 : 0), gep);
          Value* gep2 = builder.CreateConstGEP2_32(ctrlInfoArg, tranche->infoSlot, 1);
          tranche->valueMap[gep2] = gep2;
          if (!isa<ReturnInst>(termInst)) {
            int ctrlVarIdx = 0;
            for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
                 phiIterE = controlVariants.end();
                 phiIter != phiIterE; ++phiIter) {
              PHINode* phi = *phiIter;
              Value* var = phi->getIncomingValue(
                phi->getBasicBlockIndex(oldTail));
              Value* gep3 = builder.CreateConstGEP2_32(
                gep2, 0, ctrlVarIdx++);
              tranche->valueMap[gep3] = gep3;
              builder.CreateStore(var, gep3);
            }
          }
          builder.CreateRet(getConst32(tranche->infoSlot+1));
          if (unsigned int iE = termInst->getNumSuccessors()) {
            for (unsigned int i = 0; i < iE; i++) {
              BasicBlock* oldTarget = termInst->getSuccessor(i);
              if (analysis.getHead() == oldTarget) {
                termInst->setSuccessor(i, bridgeBlock);
              }
            }
          }
          else {
            BranchInst::Create(bridgeBlock, termInst);
            termInst->eraseFromParent();
          }
          blockTranche[bridgeBlock] = tranche;
          blockTranche[tailBlock]->setValueMap(bridgeBlock, bridgeBlock);
          if (tranche->nextPeer) {
            tranche->nextPeer->attachPoints.push_back(bridgeBlock);
            tranche->nextPeer->infoSlot = tranche->infoSlot + 1;
          }
        }

        // Attach tranche to parent or preceeding peer
        if (tranche->attachPoints.empty()) {
          tranche->attachPoints.push_back(cast<BasicBlock>(
            tranche->valueMap[tranche->prehead]));
        }
        for (std::vector<BasicBlock*>::iterator blkIter =
             tranche->attachPoints.begin(), blkIterE =
             tranche->attachPoints.end(); blkIter != blkIterE; ++blkIter) {
          TerminatorInst* ap = (*blkIter)->getTerminator();
          BasicBlock* target =
            cast<BasicBlock>(tranche->valueMap[tranche->head]);
          BranchInst::Create(target, ap);
          blockTranche[ap->getParent()]->setValueMap(target, target);
          ap->eraseFromParent();
        }

        // Copy parent value map entries into children
        for (std::vector<Tranche*>::iterator tIter = tranche->children.begin(),
             tIterE = tranche->children.end(); tIter != tIterE; ++tIter) {
          (*tIter)->valueMap = tranche->valueMap;
        }
      }

      // Remap instruction operands
      for (inst_iterator instIter = inst_begin(controlFlowQuant),
           instIterE = inst_end(controlFlowQuant);
           instIter != instIterE; ++instIter) {
        Instruction* inst = &*instIter;
        RemapInstruction(inst, blockTranche[inst->getParent()]->valueMap);
      }
//errs() << *controlFlowQuant;

      // Return the required number of slots
      numCtrlInfoSlots = 1 + maxSlot;
      return controlFlowQuant;
    }

    unsigned int getNumCtrlInfoSlots()
    {
      getControlFlowQuant();
      return numCtrlInfoSlots;
    }

    GlobalVariable* getTemplateBitCode()
    {
      if (templateBitCode) {
        return templateBitCode;
      }

      // Create function type for template
      std::vector<const Type*> argTypes;
      const std::set<Value*>& invariants = analysis.getInvariants();
      for (std::set<Value*>::iterator invarIter = invariants.begin(),
           invarIterE = invariants.end(); invarIter != invarIterE; ++invarIter) {
        argTypes.push_back((*invarIter)->getType());
      }
      const std::set<Instruction*>& specVals = analysis.getSpecialiseValues();
      for (std::set<Instruction*>::iterator valIter = specVals.begin(),
           valIterE = specVals.end(); valIter != valIterE; ++valIter) {
        argTypes.push_back((*valIter)->getType());
      }
      const std::set<PHINode*>& controlVariants = analysis.getControlVariants();
      for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
           phiIterE = controlVariants.end(); phiIter != phiIterE; ++phiIter) {
        argTypes.push_back((*phiIter)->getType());
      }
      const std::set<PHINode*>& runtimeVariants = analysis.getRuntimeVariants();
      for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
           phiIterE = runtimeVariants.end(); phiIter != phiIterE; ++phiIter) {
        argTypes.push_back((*phiIter)->getType());
      }
      Module* rtMod = new Module("__autojit__", ctx);
      FunctionType* templateType =
        FunctionType::get(srcFunc.getReturnType(), argTypes, false);
      Function* templateFunc = Function::Create(
        templateType, GlobalValue::ExternalLinkage, "template", rtMod);

      // Create root tranche and entry block
      DenseMap<BasicBlock*,Tranche*> blockTranche;
      Tranche* rootTranche = new Tranche();
      rootTranche->head = analysis.getHead();
      BasicBlock* entryBlock = BasicBlock::Create(ctx, "entry", templateFunc);
      rootTranche->valueMap[entryBlock] = entryBlock;
      rootTranche->prehead = entryBlock;
      blockTranche[entryBlock] = rootTranche;
      IRBuilder<> builder(entryBlock);
      builder.CreateRetVoid();

      // Setup value map
      std::set<Value*> prune;
      std::vector<const Type*> endTypes;
      iplist<Argument>::iterator args =
        templateFunc->getArgumentList().begin();
      for (std::set<Value*>::iterator invarIter = invariants.begin(),
           invarIterE = invariants.end(); invarIter != invarIterE; ++invarIter) {
        rootTranche->valueMap[*invarIter] = &*args++;
        prune.insert(*invarIter);
      }
      for (std::set<Instruction*>::iterator valIter = specVals.begin(),
           valIterE = specVals.end(); valIter != valIterE; ++valIter) {
        rootTranche->valueMap[*valIter] = &*args++;
        prune.insert(*valIter);
      }
      for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
           phiIterE = controlVariants.end(); phiIter != phiIterE; ++phiIter) {
        rootTranche->valueMap[*phiIter] = &*args++;
        prune.insert(*phiIter);
        endTypes.push_back((*phiIter)->getType());
      }
      for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
           phiIterE = runtimeVariants.end(); phiIter != phiIterE; ++phiIter) {
        rootTranche->valueMap[*phiIter] = &*args++;
        prune.insert(*phiIter);
        endTypes.push_back((*phiIter)->getType());
      }

      // Create end marker function
      FunctionType* joinMarkType = FunctionType::get(
        srcFunc.getReturnType(), endTypes, false);
      Function* joinMark = Function::Create(
        joinMarkType, GlobalValue::ExternalLinkage, "joinmark",
        rtMod);

      // Generate template for specialisation
      std::vector<Constant*> globalFixups;
      std::deque<Tranche*> tranches;
      std::vector<std::pair<BasicBlock*,BasicBlock*> > loopCuts;
      tranches.push_front(rootTranche);
      while (!tranches.empty()) {
        std::set<BasicBlock*> trancheBody, trancheFront;
        std::set<BasicBlock*> trancheJoins, trancheTails;
        Tranche* tranche = tranches.front();
        tranches.pop_front();
        trancheFront.insert(tranche->head);

        // Find the extent of the next tranche
        do {
          for (std::set<BasicBlock*>::iterator frontIter = trancheFront.begin(),
               frontIterE = trancheFront.end(); frontIter != frontIterE;) {
            BasicBlock* block = *frontIter;
            TerminatorInst* termInst = block->getTerminator();
            trancheBody.insert(block);

            Value* condition = NULL;
            if (BranchInst* brInst = dyn_cast<BranchInst>(termInst)) {
              condition = brInst->isConditional() ?
                brInst->getCondition() : NULL;
            }
            else if (SwitchInst* swInst = dyn_cast<SwitchInst>(termInst)) {
              condition = swInst->getCondition();
            }

            if (condition && isa<Instruction>(condition) &&
                !analysis.getControlFlowValues().
                  count(cast<Instruction>(condition))) {
              trancheJoins.insert(block);
              Tranche* nextPeer = NULL;
              for (Tranche* t = tranche; t; t = t->parent) {
                if (t->nextPeer) {
                  nextPeer = t->nextPeer;
                  break;
                }
              }
              for (succ_iterator succIter = succ_begin(block),
                   succIterE = succ_end(block);
                   succIter != succIterE; ++succIter) {
                BasicBlock* succ = *succIter;
                Tranche* newTranche = new Tranche();
                newTranche->head = succ;
                newTranche->prehead = block;
                newTranche->parent = tranche;
                newTranche->nextPeer = nextPeer;
                newTranche->infoSlot = tranche->infoSlot;
                nextPeer = newTranche;
                tranche->children.push_back(newTranche);
                tranches.push_front(newTranche);
              }
            }
            else {
              if (!termInst->getNumSuccessors()) {
                trancheTails.insert(block);
              }
              else {
                for (succ_iterator succIter = succ_begin(block),
                     succIterE = succ_end(block);
                     succIter != succIterE; ++succIter) {
                  BasicBlock* succ = *succIter;
                  if (analysis.getHead() == succ) {
                    trancheTails.insert(block);
                  }
                  else if (!trancheBody.count(succ)) {
                    trancheFront.insert(succ);
                  }
                }
              }
            }

           trancheFront.erase(frontIter++);
          }
        } while (!trancheFront.empty());

        // Clone blocks
        const std::list<BasicBlock*>& bodyList = analysis.getBodyList();
        for (std::list<BasicBlock*>::const_iterator blkIter = bodyList.begin(),
             blkIterE = bodyList.end(); blkIter != blkIterE; ++blkIter) {
          BasicBlock* oldBlk = *blkIter;
          if (!trancheBody.count(oldBlk)) {
            continue;
          }
          BasicBlock* newBlk = BasicBlock::Create(
            ctx, oldBlk->hasName()?oldBlk->getName():"", templateFunc);
          tranche->valueMap[oldBlk] = newBlk;
          blockTranche[newBlk] = tranche;
          for (BasicBlock::iterator instIter = oldBlk->begin(),
               instIterE = oldBlk->end(); instIter != instIterE; ++instIter) {
            Instruction* oldInst = &*instIter;
            if (!prune.count(oldInst)) {
              Instruction* newInst = oldInst->clone();
               if (oldInst->hasName()) {
                newInst->setName(oldInst->getName());
              }

              // Prune phi nodes
              if (PHINode* phi = dyn_cast<PHINode>(newInst)) {
                for (unsigned int i=0; i<phi->getNumIncomingValues(); i++) {
                  BasicBlock* pred = phi->getIncomingBlock(i);
                  if ((!trancheBody.count(pred) || trancheJoins.count(pred)) &&
                      (tranche->prehead != pred || tranche->head != oldBlk)) {
                    phi->removeIncomingValue(i--);
                  }
                }
              }

              // Insert value map entries for globals
              for (Instruction::const_op_iterator opIter = newInst->op_begin(),
                   opIterE = newInst->op_end(); opIter != opIterE; ++opIter) {
                fixupGlobals(
                  *opIter, tranche->valueMap, globalFixups, rtMod);
              }

              newBlk->getInstList().push_back(newInst);
              tranche->valueMap[oldInst] = newInst;
            }
          }

          // Find cutting points on the control-flow loops
          const std::set<Loop*>& ctrlFlowLoops =
            analysis.getControlFlowLoops();
          Loop* loop = NULL;
          for (std::set<Loop*>::iterator loopIter = ctrlFlowLoops.begin(),
               loopIterE = ctrlFlowLoops.end(); loopIter != loopIterE;
               ++loopIter) {
            if ((*loopIter)->contains(oldBlk)) {
              loop = *loopIter;
              break;
            }
          }
          if (loop) {
            TerminatorInst* termInst = oldBlk->getTerminator();
            std::vector<std::pair<BasicBlock*,BasicBlock*> > localLoopCuts;
            bool validCuts = false;
            for (unsigned int i = 0; i < termInst->getNumSuccessors(); i++) {
              if (loop->contains(termInst->getSuccessor(i))) {
                localLoopCuts.push_back(
                  std::pair<BasicBlock*,BasicBlock*>(newBlk, termInst->getSuccessor(i)));
              }
              else {
                validCuts = true;
              }
            }
            if (validCuts) {
              loopCuts.insert(
                loopCuts.end(), localLoopCuts.begin(), localLoopCuts.end());
            }
          }
        }

        // Convert tails
        for (std::set<BasicBlock*>::iterator tailIter =
             trancheTails.begin(), tailIterE = trancheTails.end();
             tailIter != tailIterE; ++tailIter) {
          BasicBlock* oldTail = *tailIter;
          BasicBlock* tailBlock = cast<BasicBlock>(tranche->valueMap[oldTail]);
          TerminatorInst* termInst = tailBlock->getTerminator();

          // Handle tails with successor blocks
          if (unsigned int iE = termInst->getNumSuccessors()) {

            // Create exit block to tag important values
            BasicBlock* exitBlock = BasicBlock::Create(
              ctx, "exit", templateFunc);
            blockTranche[exitBlock] = tranche;
            tranche->valueMap[exitBlock] = exitBlock;
            IRBuilder<> builder(exitBlock);
            std::vector<Value*> values;
            for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
                 phiIterE = controlVariants.end(); phiIter != phiIterE;
                 ++phiIter) {
              values.push_back((*phiIter)->getIncomingValueForBlock(oldTail));
            }
            for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
                 phiIterE = runtimeVariants.end(); phiIter != phiIterE;
                 ++phiIter) {
              values.push_back((*phiIter)->getIncomingValueForBlock(oldTail));
            }
            CallInst* callInst = builder.CreateCall(
              joinMark, values.begin(), values.end());
            tranche->valueMap[callInst] = callInst;

            // Break the loop to feed into exit block
            for (unsigned int i = 0; i < iE; i++) {
              BasicBlock* oldTarget = termInst->getSuccessor(i);
              if (analysis.getHead() == oldTarget) {
                termInst->setSuccessor(i, exitBlock);
              }
            }

            // Write terminator for exit block
            termInst = builder.CreateRet(callInst);
          }

          // Annotate return instruction with slot number
          Value* slotValue = getConst32(tranche->infoSlot);
          termInst->setMetadata("autojit", MDNode::get(ctx, &slotValue, 1));
        }

        // Track slot numbers
        if (tranche->nextPeer) {
          tranche->nextPeer->infoSlot = tranche->infoSlot + 1;
        }

        // Attach child tranche to parent
        TerminatorInst* termInst = cast<BasicBlock>(
          tranche->valueMap[tranche->prehead])->getTerminator();
        BasicBlock* target = cast<BasicBlock>(
          tranche->valueMap[tranche->head]);
        blockTranche[termInst->getParent()]->setValueMap(target, target);
        if (unsigned int iE = termInst->getNumSuccessors()) {
          for (unsigned int i = 0; i < iE; i++) {
            BasicBlock* oldTarget = termInst->getSuccessor(i);
            if (tranche->head == oldTarget) {
              termInst->setSuccessor(i, target);
            }
          }
        }
        else {
          // FIXME: Shouldn't need this.
          BranchInst::Create(target, termInst);
          termInst->eraseFromParent();
        }

        // Copy parent value map entries into children
        for (std::vector<Tranche*>::iterator tIter = tranche->children.begin(),
             tIterE = tranche->children.end(); tIter != tIterE; ++tIter) {
          (*tIter)->valueMap = tranche->valueMap;
        }
      }

      // Remap instruction operands
      for (inst_iterator instIter = inst_begin(templateFunc),
           instIterE = inst_end(templateFunc);
           instIter != instIterE; ++instIter) {
        Instruction* inst = &*instIter;
        RemapInstruction(inst, blockTranche[inst->getParent()]->valueMap);
      }

      // Cut open the control-flow loops
      for (std::vector<std::pair<BasicBlock*,BasicBlock*> >::iterator
           cutIter = loopCuts.begin(), cutIterE = loopCuts.end();
           cutIter != cutIterE; ++cutIter) {
        RemoveSuccessor(cutIter->first->getTerminator(),
          GetSuccessorNumber(cutIter->first, cast<BasicBlock>(
            blockTranche[cutIter->first]->valueMap[cutIter->second])));
      }

      // Eliminate dead code
      FunctionPassManager fpm(rtMod);
      fpm.add(createCFGSimplificationPass());
      fpm.add(createInstructionCombiningPass());
      fpm.add(createAggressiveDCEPass());
      fpm.doInitialization();
      fpm.run(*templateFunc);
//errs() << *templateFunc;

      // Serialise the runtime-module into a global array as bitcode
      std::vector<unsigned char> buffer;
      BitstreamWriter writer = BitstreamWriter(buffer);
      WriteBitcodeToStream(rtMod, writer);
      buffer.push_back(0);
      std::vector<Constant*> bitcodeArrayPrecursor;
      for (std::vector<unsigned char>::iterator i = buffer.begin();
           i != buffer.end(); ++i) {
        bitcodeArrayPrecursor.push_back(getConst8(*i));
      }
      templateBitCodeLen = bitcodeArrayPrecursor.size();
      Constant* bitCodeArray = ConstantArray::get(
        ArrayType::get(tyi8, templateBitCodeLen), bitcodeArrayPrecursor);
      templateBitCode = new GlobalVariable(
        *mod,
        bitCodeArray->getType(),
        true,
        GlobalValue::ExternalLinkage,
        bitCodeArray,
        prefix("templatebitcode"));
      delete rtMod;

      // Store the runtime-module's fixups into a global array
      std::vector<Constant*> fixupArrayPrecursor;
      for (std::vector<Constant*>::iterator i = globalFixups.begin();
           i != globalFixups.end(); ++i) {
        Constant* c = *i;
        if (c->getType() != Type::getInt8PtrTy(ctx)) {
          c = ConstantExpr::getBitCast(c, Type::getInt8PtrTy(ctx));
        }
        fixupArrayPrecursor.push_back(c);
      }
      templateGlobalsCount = fixupArrayPrecursor.size();
      Constant* fixupArray = ConstantArray::get(
        ArrayType::get(Type::getInt8PtrTy(ctx), templateGlobalsCount),
        fixupArrayPrecursor);
      templateGlobals = new GlobalVariable(
        *mod,
        fixupArray->getType(),
        true,
        GlobalValue::ExternalLinkage,
        fixupArray,
        prefix("templateglobals"));

      return templateBitCode;
    }

    unsigned int getTemplateBitCodeLen()
    {
      getTemplateBitCode();
      return templateBitCodeLen;
    }

    GlobalVariable* getTemplateGlobals()
    {
      getTemplateBitCode();
      return templateGlobals;
    }

    unsigned int getTemplateGlobalsCount()
    {
      getTemplateBitCode();
      return templateGlobalsCount;
    }

    Function* getGetInt()
    {
      if (getInt) {
        return getInt;
      }
      std::vector<const Type*> argTypes;
      argTypes.push_back(tyi32);
      argTypes.push_back(Type::getInt64Ty(ctx));
      FunctionType* getIntType =
        FunctionType::get(Type::getInt8PtrTy(ctx), argTypes, false);
      getInt = Function::Create(
        getIntType, GlobalValue::ExternalLinkage, "autojit_get_integer", mod);
      return getInt;
    }

    Function* getGetPtr()
    {
      if (getPtr) {
        return getPtr;
      }
      std::vector<const Type*> argTypes;
      argTypes.push_back(Type::getInt8PtrTy(ctx));
      FunctionType* getPtrType =
        FunctionType::get(Type::getInt8PtrTy(ctx), argTypes, false);
      getPtr = Function::Create(
        getPtrType, GlobalValue::ExternalLinkage, "autojit_get_pointer", mod);
      return getPtr;
    }

    Function* getConvertInvars()
    {
      if (!convertInvars) {
        convertInvars = generateConverter(getInvarsType(), "conv_invars");
      }
      return convertInvars;
    }

    Function* getConvertSpecVals()
    {
      if (!convertSpecVals) {
        convertSpecVals = generateConverter(getSpecValsType(), "conv_specvals");
      }
      return convertSpecVals;
    }

    Function* getConvertCtrlVars()
    {
      if (!convertCtrlVars) {
        convertCtrlVars = generateConverter(getCtrlVarsType(), "conv_ctrlvars");
      }
      return convertCtrlVars;
    }

    GlobalVariable* getEngineInfo()
    {
      if (engineInfo) {
        return engineInfo;
      }

      // Populate fields
      std::vector<Constant*> fields;
      fields.push_back(getControlFlowQuant());
      fields.push_back(getConvertInvars());
      fields.push_back(getConvertSpecVals());
      fields.push_back(getConvertCtrlVars());
      fields.push_back(getTemplateBitCode());
      fields.push_back(getConst32(getTemplateBitCodeLen()));
      fields.push_back(getTemplateGlobals());
      fields.push_back(getConst32(getTemplateGlobalsCount()));
      fields.push_back(getConst32(getInvarsType()->getNumElements()));
      fields.push_back(getConst32(
        targetData.getTypeStoreSize(getSpecValsType())));
      fields.push_back(getConst32(getSpecValsType()->getNumElements()));
      fields.push_back(getConst32(
        targetData.getTypeStoreSize(getCtrlVarsType())));
      fields.push_back(getConst32(getCtrlVarsType()->getNumElements()));
      fields.push_back(getConst32(
        targetData.getTypeAllocSize(getCtrlInfoType())));
      Value* indicies[] = {getConst32(0), getConst32(1)};
      fields.push_back(getConst32(targetData.getIndexedOffset(
        PointerType::getUnqual(getCtrlInfoType()), indicies, 2)));
      fields.push_back(getConst32(getNumCtrlInfoSlots()));

      // Create structure type from field types
      std::vector<const Type*> types;
      for (std::vector<Constant*>::iterator fieldIter = fields.begin(),
           fieldIterE = fields.end(); fieldIter != fieldIterE; ++fieldIter) {
        types.push_back((*fieldIter)->getType());
      }
      StructType* type = StructType::get(ctx, types, false);

      // Create global variable
      Constant* constant = ConstantStruct::get(type, fields);
      engineInfo = new GlobalVariable(
        *mod, type, true, GlobalValue::ExternalLinkage,
        constant, prefix("enginfo"));
      return engineInfo;
    }

    Function* getExternalEngine()
    {
      if (externalEngine) {
        return externalEngine;
      }
      std::vector<const Type*> types;
      types.push_back(getEngineInfo()->getType());
      types.push_back(PointerType::getUnqual(getInvarsType()));
      types.push_back(PointerType::getUnqual(getCtrlVarsType()));
      types.push_back(PointerType::getUnqual(getRtStateType()));
      FunctionType* type = FunctionType::get(
        tyi8, types, false);
      externalEngine = Function::Create(
        type, GlobalValue::ExternalLinkage, "autojit_external_engine", mod);
      return externalEngine;
    }
  };

  bool processFunction(Function& srcFunc, AnalysisPass& analysis, Pass* pass)
  {
    LLVMContext& ctx = srcFunc.getContext();
    Twine prefix = srcFunc.getName()+"__autojit__";

    FuncState state(srcFunc, *pass, analysis);
    state.getEngineInfo();

    // Remove annotations
    const std::set<CallInst*>& annotations = analysis.getAnnotations();
    for (std::set<CallInst*>::iterator annIter = annotations.begin(),
         annIterE = annotations.end(); annIter != annIterE; ++annIter) {
      CallInst* callInst = *annIter;
      callInst->replaceAllUsesWith(callInst->getOperand(1));
      callInst->eraseFromParent();
    }

    // Allocate stack space for later use
    IRBuilder<> allocaBuilder(
      &srcFunc.getEntryBlock(), srcFunc.getEntryBlock().begin());
    AllocaInst* invarsAlloca =
      allocaBuilder.CreateAlloca(state.getInvarsType());
    AllocaInst* ctrlVarsAlloca =
      allocaBuilder.CreateAlloca(state.getCtrlVarsType());
    AllocaInst* rtStateAlloca =
      allocaBuilder.CreateAlloca(state.getRtStateType());

    // Modify original function to call the external JIT engine
    BasicBlock* head = analysis.getHead();
    BasicBlock* oldHead = SplitBlock(head, head->getFirstNonPHI(), pass);
    head->getTerminator()->eraseFromParent();
    IRBuilder<> builder(head);
    unsigned int invarsIndex = 0;
    const std::set<Value*>& invariants = analysis.getInvariants();
    for (std::set<Value*>::iterator invarIter = invariants.begin(),
         invarIterE = invariants.end(); invarIter != invarIterE; ++invarIter) {
      builder.CreateStore(*invarIter,
        builder.CreateStructGEP(invarsAlloca, invarsIndex++));
    }
    unsigned int ctrlVarsIndex = 0;
    const std::set<PHINode*>& controlVariants = analysis.getControlVariants();
    for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
         phiIterE = controlVariants.end(); phiIter != phiIterE; ++phiIter) {
      builder.CreateStore(*phiIter,
        builder.CreateStructGEP(ctrlVarsAlloca, ctrlVarsIndex++));
    }
    unsigned int rtVarsIndex = 0;
    const std::set<PHINode*>& runtimeVariants = analysis.getRuntimeVariants();
    for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
         phiIterE = runtimeVariants.end(); phiIter != phiIterE; ++phiIter) {
      builder.CreateStore(*phiIter,
        buildGEP(builder, rtStateAlloca, 3, 0, 0, rtVarsIndex++));
    }
    Value* flag = builder.CreateCall4(
      state.getExternalEngine(), state.getEngineInfo(),
      invarsAlloca, ctrlVarsAlloca, rtStateAlloca);
    SwitchInst* sw = builder.CreateSwitch(flag, oldHead, 2);

    // Generate case for handling a loop continue
    BasicBlock* tailBlock = BasicBlock::Create(ctx, "tail", &srcFunc, oldHead);
    sw->addCase(state.getConst8(0), tailBlock);
    IRBuilder<> builder2(tailBlock);
    ctrlVarsIndex = 0;
    for (std::set<PHINode*>::iterator phiIter = controlVariants.begin(),
         phiIterE = controlVariants.end(); phiIter != phiIterE; ++phiIter) {
      (*phiIter)->addIncoming(builder2.CreateLoad(builder2.CreateStructGEP(
        ctrlVarsAlloca, ctrlVarsIndex++)), tailBlock);
    }
    rtVarsIndex = 0;
    for (std::set<PHINode*>::iterator phiIter = runtimeVariants.begin(),
         phiIterE = runtimeVariants.end(); phiIter != phiIterE; ++phiIter) {
      (*phiIter)->addIncoming(builder2.CreateLoad(buildGEP(
        builder2, rtStateAlloca, 3, 0, 0, rtVarsIndex++)), tailBlock);
    }
    builder2.CreateBr(head);

    // Generate case for handling a function return
    BasicBlock* retBlock = BasicBlock::Create(ctx, "ret", &srcFunc, oldHead);
    sw->addCase(state.getConst8(1), retBlock);
    IRBuilder<> builder3(retBlock);
    builder3.CreateRet(builder3.CreateLoad(builder3.CreateConstGEP2_32(
      rtStateAlloca, 0, 1)));

    return true;
  }

  bool TransformPass::runOnModule(Module& m)
  {
    bool changed = false;

    std::vector<Function*> functions;
    for (Module::iterator i = m.begin(); i != m.end(); ++i) {
      functions.push_back(&*i);
    }

    for (std::vector<Function*>::iterator i = functions.begin();
         i != functions.end(); ++i) {
      Function& f = **i;
      if (!f.isDeclaration()) {
        AnalysisPass& analysis = getAnalysis<AnalysisPass>(f);
        if (analysis.getHead()) {
          changed |= processFunction(f, analysis, this);
        }
      }
    }

    return changed;
  }

  void TransformPass::getAnalysisUsage(AnalysisUsage& aU) const
  {
    aU.addRequired<TargetData>();
    aU.addRequired<AnalysisPass>();
  }

  char TransformPass::ID = 0;
  static RegisterPass<TransformPass> X(
    "autojit",
    "AutoJIT Transformation Pass");
}
